import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd

from collections import defaultdict

from typing import Optional, Callable, Iterable, Tuple

from elrpy.data.np_utils import *
from elrpy.data.utils import *

def drop_non_numeric_cols(df: pd.DataFrame):
    """Drop non-numeric columns from dataframe.
    
    Args:
        df: Dataframe to drop columns from.

    Returns:
        pd.DataFrame: Dataframe with non-numeric columns dropped.
    """
    return df.select_dtypes(include=["number", "bool"])

def fill_na(df: pd.DataFrame):
    """Fill missing values with 0.
    
    Args:
        df: Dataframe to fill missing values in.
        
    Returns:
        pd.DataFrame: Dataframe with missing values filled with 0.
    """
    return df.fillna(0)

def drop_constant_cols(df: pd.DataFrame):
    """Drop constant columns from dataframe.
    
    Args:
        df: Dataframe to drop columns from.
    
    Returns:
        pd.DataFrame: Dataframe with constant columns dropped.
    """
    return df.loc[:, (df != df.iloc[0]).any()]


def transform_covars(
        covars: pd.DataFrame, 
        pd_transforms: Optional[Iterable[Callable]]=None
) -> np.ndarray:
    """Transform covariates into a numpy array.

    Default pandas transformations:
        drop_non_numeric_cols: Drop non-numeric columns.
        fill_na: Fill missing values with 0.
        drop_constant_cols: Drop constant columns.

    Args:
        covars (pd.DataFrame): Covariates.
        pd_transforms: List of functions to apply to covars dataframe.
        np_transforms: List of functions to apply to covars numpy array.

    Returns:
        np.ndarray: Covariates as a numpy array.
    """
    if pd_transforms is None:
        pd_transforms = [
            drop_non_numeric_cols, fill_na, drop_constant_cols
        ]
    
    for transform in pd_transforms:
        covars = transform(covars)
    covars = covars[[col for col in covars.columns if 'geo' not in col and col[:5] != 'meta_']]

    cols = covars.columns
    dd = defaultdict(list)
    for c in cols:
        if "_var_" in c:
            dd[c.split("_var_")[0]].append(c)

    X = covars.values
    X, scale = standardize(X)
    X = add_intercept(X)

    skip = [v[-2] for _, v in dd.items()]
    cc = np.array([1] + [c not in skip for i, c in enumerate(cols)]).astype(bool)
    # skip_corr = np.where(np.triu(np.abs(np.corrcoef(X.T[cc])), k=1) >= (1 - 1e-5))[0]
    # cc[cc] = np.array([True] + [(i not in skip_corr) for i, _ in enumerate(cols[cc[1:]])])
    X = X[:, cc]
    return X, covars.columns[cc[1:]], scale


def  load_from_csv(
        covars_path: str, 
        results_path: str,
        covars_group_col: Optional[str]="group_id", 
        results_group_col: Optional[str]="group_id",
        Y_cols: Optional[Iterable[str]]=["y"],
        N_cols: Optional[Iterable[str]]=["n"]
) -> Tuple[pd.Series, pd.DataFrame, pd.DataFrame]:
    """Load covariates and results from csvs and split into groups, covars and results.
    Args:
        covars_path: Path to covariates csv.
        results_path: Path to results csv.
        covars_group_col: Name of column in covars csv that contains group labels.
        results_group_col: Name of column in results csv that contains group labels.
    
    Returns:
        tuple: groups, covars, results
    """
    groups_and_covars = pd.read_csv(covars_path)
    groups = groups_and_covars[covars_group_col]
    covars = groups_and_covars.drop(columns=[covars_group_col])
    results = pd.read_csv(results_path).set_index(results_group_col)

    return groups, covars, results, Y_cols, N_cols


def  load_from_long_csv(
        covars_path: str, 
        results_path: str,
        covars_group_col: Optional[str]="group_id", 
        results_group_col: Optional[str]="group_id",
        pivot_col: Optional[str]="outcome",
        Y_col: Optional[str]="y",
        N_col: Optional[str]="n"
) -> Tuple[pd.Series, pd.DataFrame, pd.DataFrame]:
    """Load covariates and results from csvs and split into groups, covars and results.
    Args:
        covars_path: Path to covariates csv.
        results_path: Path to results csv.
        covars_group_col: Name of column in covars csv that contains group labels.
        results_group_col: Name of column in results csv that contains group labels.
        pivot_col: Name of column in results csv that contains outcome names.
        Y_col: Name of column in results csv that contains outcome values.
        N_col: Name of column in results csv that contains number of observations for the outcome.
    
    Returns:
        tuple: groups, covars, results
    """
    groups, covars, results, _, _ = load_from_csv(
        covars_path, results_path, 
        covars_group_col=covars_group_col, results_group_col=results_group_col
    )
    results = results.pivot(columns=pivot_col)
    
    Y_cols = [col for col in results.columns if Y_col in col and N_col not in col]
    N_cols = [col for col in results.columns if N_col in col]
    return groups, covars, results, Y_cols, N_cols


def transform_from_df(
        groups: pd.Series, covars: pd.DataFrame, results: pd.DataFrame,
        Y_cols: Optional[Iterable[str]]=None, 
        N_cols: Optional[Iterable[str]]=None,
        verbose: Optional[int]=1
) -> Tuple[dict, dict, dict, dict]:
    """Process csvs into dictionaries of covariates and outcomes.
    
    Args:
        groups: group labels corresponding to rows in covars.
        covars: covariates dataframe.
        results: results dataframe with index of group labels, outcomes and number of observations.
        Y_cols: Names of columns in results that contain outcomes.
        N_cols: Names of columns in results that contain number of observations.
        verbose: Verbosity level (0: silent, 1: standard prints, 2: debug prints).
        
    Returns:
        tuple: individual_frame_Xs, individual_frame_Ns, individual_frame_Ys, group_frame_Ns
    """
    if Y_cols is None:
        Y_cols = ["y"]
    if N_cols is None:
        N_cols = ["n"]

    # transform covars
    if verbose > 0:
        print("Transforming covars")
    X, columns, scale = transform_covars(covars)

    # transform to dicts
    if verbose > 0:
        print("Transforming group frame data to dicts")
    group_frame_Ys, group_frame_Ns = group_data_to_dicts(
        results.index, 
        jnp.array(results[Y_cols].values), 
        jnp.array(results[N_cols].values), 
    )
    if verbose > 0:
        print("Transforming individual frame data to dicts")
    individual_frame_Xs, individual_frame_Ns = indidual_data_to_dicts(groups, X)

    # intersect frames, dropping groups that are not in both
    if verbose > 0:
        print("Intersecting frames")
    individual_frame_Xs, individual_frame_Ns, group_frame_Ys, group_frame_Ns = intersect_frames(
        individual_frame_Xs, individual_frame_Ns, group_frame_Ys, group_frame_Ns, verbose=verbose
    )

    # get individual frame Ys by scaling group frame Ys by N_i/N_g
    individual_frame_Ys = get_individual_frame_Ys(
        individual_frame_Ns, group_frame_Ys, group_frame_Ns
    )

    return individual_frame_Xs, individual_frame_Ys, individual_frame_Ns, columns, scale
    



import jax
from jax import numpy as np

def get_wrapped_loss(loss_fn, model_fn, axis=None):
    """Wrap loss function to take model parameters and group data.

    Args:
        loss_fn: loss function.
        model_fn: model function.
        num_groups: number of groups.

    Returns:
        function: wrapped loss function.
    """
    def wrapped_loss(model_params, X, Y, G, weights=None):
        """Wrapped loss function.

        Args:
            model_params: model parameters.
            group_X: group covariates.
            group_Y: group outcomes.
            group_N: group number of observations.
            group_weight: group weights.

        Returns:
            float: loss.
        """
        return np.mean(loss_fn(
            model_fn(model_params, X), Y, G, weights=weights
        ), axis=axis)
    
    return wrapped_loss

def lyapunov_binary_loss(p, Y, G, weights=None, eps=1e-6):
    """Lyapunov Central Limit loss for binary outcomes.
    
    Args:
        p (np.ndarray): Array of probabilities of shape (n, k).
        Y (int): Number of successes.
        N (int): Size of group.
    
    Returns:
        float: a normal approximation to the combinatorial log-likelihood.
    """
    mu = G @ p
    phi2 = np.maximum(G @ (p * (1 - p)), eps)
    logp = -np.log(phi2) - (1 / phi2) * (Y - mu)**2
    if weights is not None:
        logp = weights * logp
    return -1/2 * logp



import numpy as np

def standardize(X, axis=0, eps=1e-8):
    """Standardize data along axis.

    Args:
        X: Data to standardize.
        axis: Axis along which to standardize.

    Returns:
        tuple: Standardized data and mean along axis.
    """
    # max_mask = np.max(np.abs(X), axis=axis) > (1.0 + eps)
    # mean = np.mean(X, axis=axis)
    # std = np.std(X, axis=axis)
    # print(np.max(std))
    # print(np.min(np.abs( (max_mask * std + (1 - max_mask)))))
    # return (X - mean * max_mask) / (max_mask * std + (1 - max_mask)), (mean, std, max_mask)
    max_val = np.max(np.abs(X), axis=axis)
    max_mask = max_val > (1.0 + eps)
    return X / (max_val * max_mask + (1 - max_mask)) , max_val

def add_intercept(X):
    """Add intercept to data.

    Args:
        X: Data to add intercept to.

    Returns:
        tuple: Data with intercept added and intercept.
    """
    intercept = np.ones((X.shape[0], 1))
    return np.hstack((intercept, X))

def save_transform_data(path, columns, scale):
    """Save transform data to npz file.
    
    Args:
        path: Path to save npz file.
        columns: Column names.
        scale: Scale of covariates.
    """
    np.savez(
        path, columns=columns, scale=scale
    )

def save_group_data(path, group_Xs, group_Ys, group_Ns):
    """Save group data to npz file.
    
    Args:
        path: Path to save npz file.
        group_Xs: Group covariates.
        group_Ys: Group outcomes.
        group_Ns: Group number of observations.
    """
    np.savez_compressed(
        path, group_Xs=group_Xs, group_Ys=group_Ys, group_Ns=group_Ns
    )

def load_group_data(path):
    """Load group data from npz file.
    
    Args:
        path: Path to save npz file.

    Returns:
        tuple: group_Xs, group_Ys, group_Ns
    """
    data = np.load(path, allow_pickle=True)
    return data['group_Xs'][()], data['group_Ys'][()], data['group_Ns'][()]



import jax
import jax.numpy as np

def reduce_sum(tree):
    """Reduce sum of tree.
    
    Args:
        tree: tree to sum.
        
    Returns:
        float: sum of tree.
    """
    return jax.tree_util.tree_reduce(
        jax.jit(lambda x, y: x + y if isinstance(x, np.ndarray) else [xi + yi for (xi, yi) in zip(x, y)]), 
        tree, is_leaf=lambda x: not isinstance(x, dict)
    )


def get_mapped_fn(fn):
    """Get mapped function.
    
    Args:
        fn: function to map.
            
    Returns:
        function: mapped function."""
    def mapped_fn(model_params, args):
        """Map function over tree, taking model_params as first argument and summing results.
        
        Args:
            model_params: model parameters.
            args: arguments to pass to fn.
            
        Returns:
            float: sum of mapped function.
        """
        group_values = jax.tree_util.tree_map(
            lambda *args: fn(model_params, *args), *args
        )
        return reduce_sum(group_values)
    return mapped_fn


def get_mean_fn(fn):
    def mean_fn(*args, **kwargs):
        return np.mean(fn(*args, **kwargs))
    return mean_fn

def get_dims(group_data):
    """Returns the number of groups, the number of covariates, and the number of outcomes.
    
    Args:
        group_data (tuple): tuple of group covariates, group outcomes, and group sizes
        
    Returns:
        tuple: number of groups, number of covariates, and number of outcomes
    """
    group_Xs, group_Ys, group_Ns = group_data
    num_groups = len(group_Ns)
    dim = next(iter(group_Xs.values())).shape[1]
    num_outcomes = next(iter(group_Ys.values())).shape[0]
    return num_groups, dim, num_outcomes

def get_bootstrap_weights(rng, G, num_boots, estimate_on_all=False, sample_by_group_size=True):
    """Returns bootstrap weights for each group.
    
    Args:
        rng (jax.random.PRNGKey): random number generator
        group_data (tuple): tuple of group covariates, group outcomes, and group sizes
        num_boots (int): number of bootstrap samples
        estimate_on_all (bool): whether to include replicate estimated on all data, e.g. all ones
        
    Returns:
        dict: dictionary of bootstrap weights for each group
    """
    num_boots -= estimate_on_all
    num_groups = G.shape[0]
    group_sizes = np.squeeze(G @ np.ones((G.shape[1], 1)))
    if sample_by_group_size:
        group_idx = jax.random.choice(rng, num_groups, shape=(num_groups, num_boots), p=group_sizes / np.sum(group_sizes))
    else:
        group_idx = jax.random.choice(rng, num_groups, shape=(num_groups, num_boots))
    weights = np.zeros((num_groups, num_boots))
    weights = jax.vmap(lambda w, idx: w.at[idx].add(1), in_axes=(1, 1))(weights, group_idx).T
    if estimate_on_all:
        weights = np.hstack([np.ones((num_groups, 1)), weights])
    return weights.T


import jax
import jax.numpy as np
from jax.nn import sigmoid, softmax

def binary_model(model_params, X):
    """Returns the sigmoid of the linear model.

    Args:
        model_params (np.ndarray): model parameters
        X (np.ndarray): data matrix

    Returns:
        np.ndarray: sigmoid of the linear model
    """
    p = sigmoid(X @ model_params)
    if len(p.shape) == 1:
        p = p[:, None]
    return p

def init_binary(X):
    """Initializes the model parameters for the binary model.

    Args:
        d (int): number of features
        num_outcomes (int): number of outcomes
        reps (int): number of repetitions
        rng (jax.random.PRNGKey): random seed 

    Returns:
        tuple: model and model parameters
    """
    d = X.shape[1]
    return binary_model, np.zeros((d,))

def categorical_model(model_params, X):
    """Returns the softmax of the logits of the linear model.
    Note that the last column of the logits is always zero to ensure identifiability.
    
    Args:
        model_params (np.ndarray): model parameters
        X (np.ndarray): data matrix
        
    Returns:
        np.ndarray: softmax of the logits of the linear model
    """
    model_params = model_params.reshape((X.shape[1], -1))
    
    logits = np.tensordot(X, model_params, axes=1)
    logits = np.concatenate([logits, np.zeros((*logits.shape[:-1], 1))], axis=-1)
    return softmax(logits)

def init_categorical(d, p):
    """Initializes the model parameters for the categorical model.

    Args:
        d (int): number of features
        num_outcomes (int): number of outcomes
        p (int): number of logits
        reps (int): number of repetitions
        rng (jax.random.PRNGKey): random seed
    
    Returns:
        tuple: model and model parameters
    """
    return categorical_model, np.zeros((d * (p - 1),))


import jax
from jax import numpy as np

def get_bootstrap_weights(rng, group_data, num_boots, estimate_on_all=False, sample_by_group_size=True):
    """Returns bootstrap weights for each group.
    
    Args:
        rng (jax.random.PRNGKey): random number generator
        group_data (tuple): tuple of group covariates, group outcomes, and group sizes
        num_boots (int): number of bootstrap samples
        estimate_on_all (bool): whether to include replicate estimated on all data, e.g. all ones
        
    Returns:
        dict: dictionary of bootstrap weights for each group
    """
    num_boots -= estimate_on_all
    num_groups = len(group_data[0])
    group_sizes, tree_def = jax.tree_flatten(jax.tree_map(lambda x: x.shape[0], group_data[0])) 
    group_sizes = np.array(group_sizes)
    if sample_by_group_size:
        group_idx = jax.random.choice(rng, num_groups, shape=(num_groups, num_boots), p=group_sizes / np.sum(group_sizes))
    else:
        group_idx = jax.random.choice(rng, num_groups, shape=(num_groups, num_boots))
    weights = np.zeros((num_groups, num_boots))
    weights = jax.vmap(lambda w, idx: w.at[idx].add(1), in_axes=(1, 1))(weights, group_idx).T
    if estimate_on_all:
        weights = np.hstack([np.ones((num_groups, 1)), weights])
    group_weights = {g: w[:, None] for g, w in zip(group_data[0], weights)}
    return group_weights

def bootstrap_marginal_confidence_interval(bootstrap_params, alpha=0.05):
    """Returns the confidence interval for the given bootstrap parameters.
    
    Args:
        bootstrap_params (np.ndarray): bootstrap parameters
        alpha (float): confidence level
        
    Returns:
        tuple: lower and upper confidence bounds
    """
    lower = np.quantile(bootstrap_params, 100 * (alpha / 2), axis=0)
    upper = np.quantile(bootstrap_params, 100 * (1 - alpha / 2), axis=0)
    return lower, upper

def bootstrap_prediction_confidence_interval(model_fn, bootstrap_params, X, G=None, alpha=0.05):
    """Returns the confidence interval for the given bootstrap parameters.
    
    Args:
        model_fn (function): model function
        bootstrap_params (np.ndarray): bootstrap parameters
        X (np.ndarray): data matrix
        G (np.ndarray): group matrix
        alpha (float): confidence level
        
    Returns:
        tuple: lower and upper confidence bounds
    """
    bootstrap_y = model_fn(bootstrap_params, X)
    if G is not None:
        bootstrap_y = (G @ bootstrap_y.T).T

    lower = np.quantile(bootstrap_y, 100 * (alpha / 2), axis=0)
    upper = np.quantile(bootstrap_y, 100 * (1 - alpha / 2), axis=0)
    return lower, upper


import jax
from jax import numpy as np
from typing import Any, Optional, Tuple, Callable

def get_cg_fn(loss_and_grad_fn, hess_fn, l2=0.0):
    """Return standard conjugate gradient function, using lstsq to solve the linear system in case of ill-conditioning.

    Args:
        grad_fn: function returning loss and gradient.
        hess_fn: Hessian function.
        l2: L2 regularization strength.

    Returns:
        function: conjugate gradient function.
    """
    def cg_fn(params, *args):
        """Conjugate gradient function using lstsq to solve the linear system in case of ill-conditioning.

        Args:
            params: model parameters.
            *args: arguments to pass to loss and gradient function.

        Returns:
            tuple: loss, conjugate gradient, gradient norm.
        """
        loss, grad = loss_and_grad_fn(params, *args) 
        hess = hess_fn(params, *args)
        loss += 0.5 * l2 * np.sum(params ** 2)
        grad += l2 * params
        hess += l2 * np.eye(hess.shape[-1])
        return np.mean(loss), np.linalg.lstsq(hess, grad)[0], np.linalg.norm(grad, axis=-1)
    return cg_fn

def clip_inv_prod(eps, grad, u, v):
    """Clip eigenvalues and compute inverse product.
    
    Args:
        eps: clipping values.
        grad: gradient.
        u: eigenvalues.
        v: eigenvectors.
        
    Returns:
        np.ndarray: inverse product.
    """
    inv_hess = v @ np.diag(1 / np.maximum(u, eps)) @ v.T
    next_cg = inv_hess @ grad
    return next_cg

eigh =  jax.jit(np.linalg.eigh)
clip_inv_prod = jax.jit(jax.vmap(clip_inv_prod, in_axes=(0, None, None, None)))

def get_clipped_cg_fn(loss_fn, grad_fn, hess_fn, start_clip=1e-1, stop_clip=1e-4, num_clip=10, l2=0.0):
    """Return conjugate gradient function with eigenvalue clipping in case of ill-conditioning.
    Eigenvalues are clipped between start_clip and stop_clip, with num_clip values in between,
    with the clipped inverse product computed using the eigendecomposition of the Hessian.
    The conjugate gradient is then computed as the minimum of the clipped inverse products and the
    standard conjugate gradient.

    Args:
        loss_fn: loss function.
        grad_fn: gradient function.
        hess_fn: Hessian function.
        start_clip: starting clipping value.
        stop_clip: stopping clipping value.
        num_clip: number of clipping values.
        l2: L2 regularization strength.

    Returns:
        function: conjugate gradient function.
    """

    # compute clipping values
    eps = np.logspace(
        np.log(start_clip), np.log(stop_clip), num_clip, base=np.exp(1)
    )

    # define vectorizing helper functions
    def cp(params, grad, hess):
        """Compute conjugate gradient using clipped inverse product.
        
        Args:
            params: model parameters.
            grad: gradient.
            hess: Hessian.
            
        Returns:
            tuple: gradient, Hessian, updated parameters, conjugate gradient.
        """
        grad += l2 * params
        hess += l2 * np.eye(hess.shape[0])
        u, v = eigh(hess)
        cg = clip_inv_prod(eps, grad, u, v).T
        return grad, hess, params[:, None] - cg, cg
    
    def gl(losses, cg, params, grad):
        """Compute loss, conjugate gradient and gradient norm.

        Args:
            losses: losses.
            cg: conjugate gradient.
            params: model parameters.
            grad: gradient.

        Returns:
            tuple: losses, conjugate gradient, gradient norm.
        """
        min_idx = np.argmin(losses + 0.5 * l2 * np.sum((params[:, None] - cg) ** 2, axis=0), axis=0)
        return losses[min_idx], cg[:, min_idx], np.linalg.norm(grad)

    # vectorize helper functions
    # cp = jax.jit(jax.vmap(cp, in_axes=(0, 0, 0)))
    # gl = jax.jit(jax.vmap(gl, in_axes=(0, 0, 0, 0)))

    def cg_fn(params, *args):
        """Conjugate gradient function using clipped inverse product.

        Args:
            params: model parameters.
            *args: arguments to pass to loss and gradient function.

        Returns:
            tuple: loss, conjugate gradient, gradient norm.
        """
        loss, grad = grad_fn(params, *args) 
        hess = hess_fn(params, *args)
        grad, hess, updated_params, cg = cp(params, grad, hess)
        losses = jax.vmap(loss_fn, in_axes=(1, *([None] * len(args))))(updated_params, *args) 
        losses, cg, grad_norm = gl(losses, cg, params, grad)
        return loss, cg, np.max(grad_norm)
    return cg_fn

def fit(
        loss_fn: Callable, 
        model_fn: Callable,
        model_params: dict, 
        group_data: Tuple[dict],
        lr: Optional[float] = 1.0,
        maxit: Optional[int] = 1_000, 
        tol: Optional[float] = 1e-8, 
        verbose: Optional[int] = 0, 
        print_every: Optional[int] = 50, 
        save_every: Optional[int] = 50,
        save_dir: Optional[str] = None,
        mapped_loss_and_dir_fn = None,
        group_weights=None,
        keep_history=False,
        dir_type='cg'
) -> Tuple[dict, float]:
    """Fit a model to data using directional descent with the given loss and direction functions.

    Args:
        loss_fn: loss function.
        model_fn: model function.
        model_params: model parameters.
        group_data: tuple of group covariates, group outcomes and group number of observations.
        maxit: maximum number of iterations.
        tol: tolerance for convergence.
        verbose: verbosity level (0: silent, 1: standard prints, 2: debug prints).
        print_every: print every n iterations.
        save_every: save model every n iterations.
        save_dir: directory to save model.
        lr: learning rate.
        mapped_loss_and_dir_fn: mapped loss and direction function, if None, will be compiled from loss function.
        group_weights: weights for each group.
        keep_history: whether to keep history of model parameters, loss and gradient norm.

    Returns:
        tuple: model parameters, gradient norm.
    """
    if mapped_loss_and_dir_fn is None:
        from elrpy.losses import get_wrapped_loss
        if verbose > 0:
            print("Gradient functions not provided, these will be recompiled...")
        wrapped_loss_fn = get_wrapped_loss(loss_fn, model_fn)
        if dir_type == 'cg':
            if group_weights is None:
                _loss_fn = wrapped_loss_fn
                loss_and_grad_fn = jax.value_and_grad(wrapped_loss_fn)
                hess_fn = jax.hessian(wrapped_loss_fn)
            else:
                _loss_fn = jax.vmap(wrapped_loss_fn, in_axes=(0, None, None, None))
                loss_and_grad_fn = jax.vmap(jax.value_and_grad(wrapped_loss_fn), in_axes=(0, None, None, None))
                hess_fn = jax.vmap(jax.hessian(wrapped_loss_fn), in_axes=(0, None, None, None))
            mapped_loss_and_dir_fn = jax.jit(get_clipped_cg_fn(_loss_fn, loss_and_grad_fn, hess_fn))
        elif dir_type == 'gd':
            mapped_loss_and_dir_fn = jax.jit(jax.value_and_grad(wrapped_loss_fn))

    if group_weights is not None:
        group_data = (group_data[0], group_data[1], group_data[2], group_weights)

    if keep_history:
        history = []

    for i in range(maxit):
        out = mapped_loss_and_dir_fn(model_params, *group_data)
        if len(out) == 2:
            loss, grad = out
            grad_norm = np.linalg.norm(grad, axis=-1)
        elif len(out) == 3:
            loss, grad, grad_norm = out
        else:
            raise ValueError("Unexpected number of outputs from mapped loss and gradient function.")
        
        if keep_history:
            history.append((model_params, loss, grad_norm))

        if np.any(np.isnan(grad)):
            if verbose > 0:
                print(f"NaN gradient update, aborting with loss {loss}...")
            break
        
        if i % print_every == 0 and verbose == 2:
            print(i, "\t", loss, "\t", grad_norm)
        if np.all(grad_norm < tol) and verbose > 0:
            print("Converged!")
            if verbose == 2:
                print(i, loss, grad_norm)
            break

        model_params -= (grad_norm >= tol) * lr * grad
        if i % save_every == 0 and save_dir is not None:
            np.savez(f"{save_dir}/model.npz", model_params=model_params, grad_norm=grad_norm, i=i)

    if grad_norm > tol and verbose > 0:
        print(f"Failed to converge, gradient norm is {grad_norm}.")
    if save_dir is not None:
        np.savez(f"{save_dir}/model.npz", model_params=model_params, grad_norm=grad_norm, i=i)

    if keep_history:
        params, losses, grad_norms = (np.array(z) for z in zip(*history))
        history = (params, losses, grad_norms)
    
    return (model_params, grad_norm, history) if keep_history else (model_params, grad_norm)




def normal_marginal_confidence_interval(grad_fn, hess_fn, params, alpha=0.05):
    """Returns the confidence interval for the given parameters.
    
    Args:
        grad_fn (function): gradient function
        hess_fn (function): Hessian function
        params (np.ndarray): parameters
        alpha (float): confidence level
        
    Returns:
        tuple: lower and upper confidence bounds
    """
    grad = grad_fn(params)
    hess = hess_fn(params)
    z = np.sqrt(params.shape[0]) * np.linalg.solve(hess, grad)
    lower = params - norm.ppf(1 - alpha / 2) * np.sqrt(np.diag(np.linalg.inv(hess)))
    upper = params + norm.ppf(1 - alpha / 2) * np.sqrt(np.diag(np.linalg.inv(hess)))
    


import jax
import numpy as np

from typing import Iterable, Tuple, Optional


def group_data_to_dicts(
        groups: Iterable, Y: Iterable, N: Iterable
) -> Tuple[dict, dict]: 
    """Convert group data to dictionaries. All inputs must be the same length.
    
    Args:
        groups: Group labels.
        Y: Outcomes.
        N: Number of observations per group.
        
    Returns:
        tuple: dict, group_frame_Ns
    """
    group_frame_Ys = dict(zip(groups, Y))
    group_frame_Ns = dict(zip(groups, N))
    return group_frame_Ys, group_frame_Ns

def indidual_data_to_dicts(groups: np.ndarray, X: np.ndarray) -> Tuple[dict, dict]:
    group_indices = {
        group: np.where(group == groups)[0] for group in np.unique(groups)
    }
    individual_frame_Xs = jax.tree_util.tree_map(lambda idx: X[idx], group_indices)
    individual_frame_Ns = jax.tree_util.tree_map(lambda idx: idx.shape[0], group_indices)
    return individual_frame_Xs, individual_frame_Ns

def intersect_frames(
        individual_frame_Xs: dict, individual_frame_Ns: dict,
        group_frame_Ys: dict, group_frame_Ns: dict, 
        verbose: Optional[int] = 1
) -> Tuple[dict, dict, dict, dict]:
    """Remove groups that are not in both individual and group frame.
    
    Args:
        individual_frame_Xs: Individual frame covariates.
        individual_frame_Ns: Individual frame number of observations.
        group_frame_Ys: Group frame outcomes.
        group_frame_Ns: Group frame number of observations.
        verbose: Verbosity level (0: silent, 1: number of groups removed, 2: names of groups removed).
        
    Returns:
        tuple: individual_frame_Xs, individual_frame_Ns, group_frame_Ys, group_frame_Ns
    """
    individual_frame_groups = set(individual_frame_Xs.keys())
    group_frame_groups = set(group_frame_Ys.keys())
    # remove groups that are not in both individual and group frame
    n_dropped_individual_frame = 0
    not_in_group_frame = individual_frame_groups.difference(group_frame_groups)
    for g in not_in_group_frame:
        if verbose == 2:
            print(f"Dropping group {g} because it is not in the group frame")
        n_dropped_individual_frame += np.max(individual_frame_Ns[g])
        individual_frame_Xs.pop(g)
        individual_frame_Ns.pop(g)
    
    # remove groups that are not in both individual and group frame
    n_dropped_group_frame = 0
    not_in_individual_frame = group_frame_groups.difference(individual_frame_groups)
    for g in not_in_individual_frame:
        if verbose == 2:
            print(f"Dropping group {g} because it is not in the individual frame")
        n_dropped_group_frame += np.max(group_frame_Ns[g])
        group_frame_Ys.pop(g)
        group_frame_Ns.pop(g)
    
    if verbose > 0:
        print(
            f"Dropped {len(not_in_group_frame)} groups with "
            f"{n_dropped_individual_frame} individuals from individual frame."
        )
        print(
            f"Dropped {len(not_in_individual_frame)} groups with "
            f"{n_dropped_group_frame} individual outcomes from group frame"
        )
    
    return individual_frame_Xs, individual_frame_Ns, group_frame_Ys, group_frame_Ns

def get_individual_frame_Ys(
        individual_frame_Ns: dict, group_frame_Ys: dict, group_frame_Ns: dict
) -> dict:
    """Normalize group frame outcomes to individual frame.
    
    Args:
        individual_frame_Ns: Individual frame number of observations.
        group_frame_Ys: Group frame outcomes.
        group_frame_Ns: Group frame number of observations.
        
    Returns:
        dict: Individual frame outcomes."""
    individual_frame_Ys = jax.tree_util.tree_map(
        lambda N_i, Y_g, N_g: Y_g * N_i / N_g,
        individual_frame_Ns, group_frame_Ys, group_frame_Ns
    )
    return individual_frame_Ys


import os
from elrpy.data.pd_utils import load_from_csv, transform_from_df
from elrpy.data.np_utils import save_transform_data, save_group_data, load_group_data

from typing import Optional, Iterable, Tuple 

def load(
    covars_path: str, 
    results_path: str,
    covars_group_col: Optional[str]="group_id", 
    results_group_col: Optional[str]="group_id",
    Y_cols: Optional[str]=None, 
    N_cols: Optional[str]=None,
    save_npz: Optional[bool]=True,
    force_from_csv: Optional[bool]=False,
    verbose: Optional[bool]=1
) -> Tuple[dict, dict, dict]:
    """Load covariates and results from csvs and split into groups, covars and results.

    Args:
        covars_path: Path to covariates csv.
        results_path: Path to results csv.
        covars_group_col: Name of column in covars csv that contains group labels.
        results_group_col: Name of column in results csv that contains group labels.
        Y_cols: Names of columns in results that contain outcomes.
        N_cols: Names of columns in results that contain number of observations.
        save_npz: Whether to save npz file of group data.
        force_from_csv: Whether to force loading from csvs.
        verbose: Verbosity level (0: silent, 1: standard prints, 2: debug prints).

    Returns:
        tuple: individual_frame_Xs, individual_frame_Ns, individual_frame_Ys, group_frame_Ns
    """
    
    data_dir = os.path.dirname(results_path)
    covars_name = os.path.basename(covars_path).split(".")[0]
    results_name = os.path.basename(results_path).split(".")[0]
    npz_path = os.path.join(data_dir, f"{results_name}.{covars_name}.npz")
    if os.path.exists(npz_path) and not force_from_csv:
        if verbose > 0:
            print(f"Loading data from {npz_path}.\n"
                  "\tSet force_from_csv=True to reload the data using updated files or group_cols/Y_cols/N_cols.")
        return load_group_data(npz_path)
    
    if verbose > 0:
        print("Loading data from csvs")

    groups, covars, results, Y_cols, N_cols = load_from_csv(
        covars_path, results_path, 
        covars_group_col=covars_group_col, results_group_col=results_group_col,
        Y_cols=Y_cols, N_cols=N_cols
    )

    if verbose > 0:
        print("Transforming data from csvs.")
    group_Xs, group_Ys, group_Ns, columns, scale = transform_from_df(
        groups, covars, results, Y_cols=Y_cols, N_cols=N_cols, verbose=verbose
    )

    if save_npz:
        if verbose > 0:
            print("Saving data to npz.")
        save_transform_data(os.path.join(os.path.dirname(covars_path), f"transforms.npz"), columns, scale)
        save_group_data(npz_path, group_Xs, group_Ys, group_Ns)
    
    return group_Xs, group_Ys, group_Ns




import jax
from jax import numpy as np
from jax.nn import sigmoid

from jax.experimental import sparse
def normal_sim_binary(rng, n, d, k, beta=None, gamma=None, intercept=True):
    """Simulate data from a normal mixture model with binary outcomes.

    Args:
        rng (jax.random.PRNGKey): random number generator
        n (int): number of observations
        d (int): number of features
        k (int): number of groups
        beta (np.ndarray): regression coefficients
        gamma (np.ndarray): group coefficients
    
    Returns:
        np.ndarray: data matrix
        np.ndarray: outcome vector
        dict: group indices
        dict: group data matrices
        dict: group outcome vectors
    """
    rng, next_rng = jax.random.split(rng)
    if intercept:
        X = jax.random.normal(next_rng, shape=(n, d - 1))
        X = np.hstack([np.ones((n, 1)), X])
    else:
        X = jax.random.normal(next_rng, shape=(n, d))

    if beta is None:
        rng, next_rng = jax.random.split(rng)
        beta = jax.random.normal(next_rng, shape=(d,))

    if gamma is None:
        rng, next_rng = jax.random.split(rng)
        gamma = jax.random.normal(next_rng, shape=(d, k))

    rng, next_rng = jax.random.split(rng)
    g = jax.random.categorical(rng, X @ gamma)
    G = sparse.BCOO((np.ones(g.shape[0]), np.vstack([np.arange(g.shape[0]), g]).T), shape=(g.shape[0], k)).T

    p = sigmoid(X @ beta)
    rng, next_rng = jax.random.split(rng)
    Y = jax.random.bernoulli(next_rng, p)

    return beta, gamma, X, Y, G @ Y, G


def normal_sim_categorical(rng, n, d, k, p, eps=1e-6, beta=None, gamma=None):
    """Simulate data from a normal mixture model with categorical outcomes.
    
    Args:
        rng (jax.random.PRNGKey): random number generator
        n (int): number of observations
        d (int): number of features
        k (int): number of groups
        p (int): number of outcomes
        eps (float): small constant to avoid numerical issues
        beta (np.ndarray): regression coefficients
        gamma (np.ndarray): group coefficients

    Returns:
        np.ndarray: data matrix
        np.ndarray: outcome vector
        dict: group indices
        dict: group data matrices
        dict: group outcome vectors
    """
    rng, next_rng = jax.random.split(rng)
    X = jax.random.normal(next_rng, shape=(n, d - 1))
    X = np.hstack([np.ones((n, 1)), X])

    if beta is None:
        rng, next_rng = jax.random.split(rng)
        beta = jax.random.normal(next_rng, shape=(d, p - 1))
        beta = np.concatenate([beta, np.zeros((d, 1))], axis=-1)

    if gamma is None:
        rng, next_rng = jax.random.split(rng)
        gamma = jax.random.normal(next_rng, shape=(d, k))

    rng, next_rng = jax.random.split(rng)
    g = jax.random.categorical(rng, X @ gamma)
    G = jax.nn.one_hot(g, k)
    G = sparse.BCOO.fromdense(G).T

    rng, next_rng = jax.random.split(rng)
    Y = jax.random.categorical(rng, np.tensordot(X, beta, axes=1))
    Y = jax.nn.one_hot(Y, p)
    return beta.flatten(), gamma, X, Y, G @ Y, G